import { Route, APIRoute, PageRoute, MiddlewareRoute, RouteService, Method } from "../../route"
import { ExpressService } from "../expressService"
import { UserService } from "../userService"
import { UserDefinitionService, User } from "../userDefinitionService"
import { RouteSessionService } from "../routeSessionService"
import * as express from "express"
import * as Leaf from "leaf-ts"

declare const module
export const Errors = Leaf.ErrorDoc.create()
    .define("Forbidden")
    .define("AuthorizationFailed")
    .generate() as Leaf.ErrorsOf<{
        Forbidden: {}
        AuthorizationFailed: {}
    }>
export class UserAPIService<TUser extends User = User> extends RouteService {
    readonly name = "api.user"
    dependencies = ["ExpressService", "RouteSessionService", "UserService", "UserDefinitionService"]
    prefix = "/api/"
    services: {
        ExpressService: ExpressService
        RouteSessionService: RouteSessionService
        UserService: UserService<TUser>
        UserDefinitionService: UserDefinitionService<TUser>
    }
    initialize() {
        this.addRoutes(module.exports)
        this.installTo(this.services.ExpressService.server)
    }
}

export class SessionUserProvider extends MiddlewareRoute {
    services: {
        UserService: UserService<User>
        RouteSessionService: RouteSessionService
    }
    req: express.Request & { session: { username: string } }
    info: { user: User }
    handle() {
        let session = this.req.session
        if (!session) {
            this.next()
        }
        let name = this.req.session.username
        if (!name) {
            this.next()
            return
        }
        this.services.UserService.getUserByUsername(name, (err, user) => {
            if (err) {
                this.next()
                return
            }
            this.info.user = user
            this.next()
        })
    }
}

export class Register extends APIRoute<null> {
    readonly method = Method.POST
    readonly path = "user"
    readonly minUsernameLength = 3
    readonly maxUsernameLength = 32
    readonly minPasswordLength = 8
    readonly maxPasswordLength = 1024
    services: {
        UserService: UserService<User>
        RouteSessionService: RouteSessionService
        UserDefinitionService: UserDefinitionService<User>
    }
    body: Partial<User>
    private checkUserValidation(user: Partial<User>) {
        return this.services.UserDefinitionService.UserValidator.check(user)
    }
    req: express.Request & { session: { username: string } }
    info: { user: User }
    handle() {
        let userToCheck = this.body
        try {
            this.checkUserValidation(userToCheck)
        } catch (err) {
            this.error(err);
            return
        }
        this.services.UserService.createUser(userToCheck as User, (err, user) => {
            if (err) {
                this.error(err)
                return
            }
            this.req.session.username = user.username
            this.success(null)
        })
    }
}

export class GetUserProfile extends APIRoute<User>{
    readonly method = Method.GET
    readonly path = "user"
    services: {
        UserService: UserService<User>
    }
    info: {
        user: User
    }
    handle() {
        if (!this.info.user) {
            this.error(new Errors.Forbidden)
            return
        }
        let user = this.info.user
        this.success({
            username: user.username,
            email: user.email,
            displayName: user.displayName
        })
    }
}

export class Login extends APIRoute<User>{
    readonly method = Method.PUT
    readonly path = "user"
    services: {
        UserService: UserService<User>
    }
    body: {
        login: string
        password: string
    }
    req: express.Request & { session: { username: string } }
    handle() {
        let loginUser = this.body
        this.services.UserService.getUserByLogin(loginUser, (err, user) => {
            if (err) {
                this.error(err)
                return
            }
            if (!user) {
                this.error(new Errors.AuthorizationFailed)
                return
            }
            this.req.session.username = user.username
            this.success(user)
        })
    }
}

export class Logout extends APIRoute<null> {
    readonly method = Method.DELETE
    readonly path = "user"
    req: express.Request & { session: { username: string } }
    services: {
        UserService: UserService<User>
    }
    handle() {
        this.req.session.username = null
        this.success(null)
    }
}
