import { Service } from "../service"
import * as Leaf from "leaf-ts"

export const Errors = Leaf.ErrorDoc.create()
    .define("InvalidUsername")
    .define("UsernameTooShort")
    .define("UsernameTooLong")
    .define("InvalidUsername")
    .define("PasswordTooShort")
    .define("PasswordTooLong")
    .define("InvalidPassword")
    .define("InvalidEmail")
    .define("EmailTooLong")
    .define("InvalidDisplayName")
    .define("DisplayNameTooShort")
    .define("DisplayNameTooLong")
    .define("UsernameRequired")
    .define("PasswordRequired")
    .define("EmailRequired")
    .generate() as Leaf.ErrorsOf<{
        InvalidUsername: {}
        UsernameTooShort: {}
        UsernameTooLong: {}
        PasswordTooShort: {}
        PasswordTooLong: {}
        EmailTooLong: {}
        InvalidEmail: {}
        InvalidDisplayName: {}
        DisplayNameTooShort: {}
        DisplayNameTooLong: {}
        InvalidPassword: {}
        UsernameRequired: {}
        PasswordRequired: {}
        EmailRequired: {}
    }>
export type User = UserDefinitionService.User
export class UserDefinitionService<TUser extends UserDefinitionService.User> extends Service {
    readonly name = "UserDefinitionService"
    public UserPartialValidator: Leaf.Validator = GenerateUserValidator()
    public UserValidator = this.UserPartialValidator.clone()
        .field("username").exists().error(new Errors.UsernameRequired)
        .field("email").exists().error(new Errors.EmailRequired)
        .field("password").exists().error(new Errors.PasswordRequired)
}
export namespace UserDefinitionService {
    export interface User {
        id?: string
        username: string
        displayName?: string
        passwordHash?: string
        password?: string
        email: string
        createAt?: number
    }
}

export interface UserValidationConfig {
    minUsernameLength: number
    maxUsernameLength: number
    minDisplayNameLength: number
    maxDisplayNameLength: number
    minPasswordLength: number
    maxPasswordLength: number
    maxEmailLength: number
}
export const DefaultUserValidationConfig: UserValidationConfig = {
    minUsernameLength: 3,
    maxUsernameLength: 32,
    minDisplayNameLength: 3,
    maxDisplayNameLength: 32,
    minPasswordLength: 8,
    maxPasswordLength: 256,
    maxEmailLength: 128
}
export function GenerateUserValidator(config: UserValidationConfig = DefaultUserValidationConfig) {
    return Leaf.Validator.create()
        .field("username")
        .string().error(new Errors.InvalidUsername)
        .gt(config.minUsernameLength - 1).error(new Errors.UsernameTooShort)
        .lt(config.maxUsernameLength + 1).error(new Errors.UsernameTooLong)
        .match(/^[0-9a-z_]+$/i).error(new Errors.InvalidUsername)

        .field("displayName")
        .string().error(new Errors.InvalidDisplayName)
        .gt(config.minDisplayNameLength - 1).error(new Errors.DisplayNameTooShort)
        .lt(config.maxDisplayNameLength + 1).error(new Errors.DisplayNameTooLong)

        .field("password")
        .string().error(new Errors.InvalidPassword)
        .gt(config.minPasswordLength - 1).error(new Errors.PasswordTooShort)
        .lt(config.maxPasswordLength + 1).error(new Errors.PasswordTooLong)

        .field("email")
        .lt(config.maxEmailLength + 1).error(new Errors.EmailTooLong)
        .match(/^(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/)
        .readonly()
}
